package kr.pe.kwonnam.reflectioninjector;

import static kr.pe.kwonnam.reflectioninjector.ReflectionInjectorUtils.*;
import static org.hamcrest.CoreMatchers.*;
import static org.junit.Assert.*;

import java.util.Date;

import org.junit.Before;
import org.junit.Test;

public class ReflectionInjectorUtilsTest {

	private TestServiceImpl testService;

	private TestDaoImpl testDaoImpl;

	private TestDaoImpl testDaoAnother;

	private TestDaoSecondImpl testDaoSecondImpl;

	private TestConcreteClassDao testConcreteClassDao;

	private TestParentDao testParentDao;

	private ReflectionInjector injector;

	@Before
	public void setUp() {
		testService = new TestServiceImpl();
		testDaoImpl = new TestDaoImpl();
		testDaoAnother = new TestDaoImpl();
		testDaoSecondImpl = new TestDaoSecondImpl();
		testConcreteClassDao = new TestConcreteClassDao();
		testParentDao = new TestParentDao();

		injector = injector(testService);
	}

	@Test
	public void constructor() {
		assertThat((TestServiceImpl) injector.getTarget(),
				sameInstance(testService));
	}

	@Test(expected = IllegalArgumentException.class)
	public void set_주입대상없음() {
		Date notExistingFieldObject = new Date();
		injector.set(notExistingFieldObject);
	}

	@Test
	public void set_자동판단주입() {
		ReflectionInjector ri = injector.set(testConcreteClassDao);

		assertThat(ri, sameInstance(injector));
		assertThat(testService.getTestConcreteClass(),
				sameInstance(testConcreteClassDao));
	}

	@Test(expected = IllegalArgumentException.class)
	public void set_지정클래스판단주입_주입대상없음() {
		Date notExistingField = new Date();
		injector.set(Date.class, notExistingField);
	}

	@Test(expected = IllegalArgumentException.class)
	public void set_지정클래스판단주입_주입할_값이_지정된_클래스를_상속하지_않음() {
		Date notExisingField = new Date();
		injector.set(TestDao.class, notExisingField);
	}

	@Test
	public void set_지정클래스판단주입() {
		ReflectionInjector ri = injector.set(TestDaoSecondImpl.class,
				testDaoSecondImpl);

		assertThat(ri, sameInstance(injector));

		assertThat(testService.getTestDao(), nullValue());
		assertThat(testService.getTestDaoAnother(), nullValue());
		assertThat(testService.getTestDaoSecondImpl(),
				sameInstance(testDaoSecondImpl));
	}

	@Test(expected = IllegalArgumentException.class)
	public void set_필드명지정주입_주입대상없음() {
		injector.set("notExistingFieldName", testDaoImpl);
	}

	@Test(expected = IllegalArgumentException.class)
	public void set_필드명지정주입_주입대상은있으나_구현클래스가다름() {
		injector.set("testDao", testConcreteClassDao);
	}

	@Test
	public void set_필드명지정주입() {
		ReflectionInjector ri = injector.set("testDao", testDaoImpl);
		assertThat(ri, sameInstance(injector));

		assertThat(testService.getTestDao(),
				sameInstance((TestDao) testDaoImpl));
		assertThat(testService.getTestDaoAnother(), nullValue());
		assertThat(testService.getTestDaoSecondImpl(), nullValue());
	}

	@Test
	public void set_필드명지정주입_another() {
		injector.set("testDaoAnother", testDaoAnother);

		assertThat(testService.getTestDaoAnother(),
				sameInstance((TestDao) testDaoAnother));
		assertThat(testService.getTestDao(), nullValue());
		assertThat(testService.getTestDaoSecondImpl(), nullValue());
	}

	@Test
	public void get_null() {
		injectAll();

		Date result = injector.get(Date.class, "date");
		assertThat(result, nullValue());

		result = injector.get(Date.class);
		assertThat(result, nullValue());
	}

	private void injectAll() {
		injector.set("testDao", testDaoImpl);
		injector.set("testDaoAnother", testDaoAnother);
		injector.set(TestDaoSecondImpl.class, testDaoSecondImpl);
		injector.set(testConcreteClassDao);
		injector.set(testParentDao);
	}

	@Test(expected = ClassCastException.class)
	public void get_존재하는_필드명이나_클래스타입_불일치() {
		injectAll();

		injector.get(TestDao.class, "testConcreteClass");
	}

	@Test(expected = ClassCastException.class)
	public void get_존재하는_필드명이나_클래스타입_불일치_필드가_더_상위클래스인경우() {
		injectAll();

		injector.get(TestDaoSecondImpl.class, "testDao");
	}

	@Test
	public void get_class() {
		injectAll();

		TestDaoSecondImpl result = injector.get(TestDaoSecondImpl.class);
		assertThat(result, sameInstance(testDaoSecondImpl));
	}

	@Test
	@SuppressWarnings("unchecked")
	public void get_class_중복있을경우() {
		injectAll();

		TestDao result = injector.get(TestDao.class);

		assertThat((TestDaoImpl) result,
				anyOf(sameInstance(testDaoImpl), sameInstance(testDaoAnother)));
		assertThat(result, not(sameInstance((TestDao) testDaoSecondImpl)));
	}

	@Test
	public void get_필드명() {
		injectAll();

		TestDao result = injector.get(TestDao.class, "testDao");

		assertThat(result, sameInstance((TestDao) testDaoImpl));
	}

	@Test
	public void get_필드명_another() {
		injectAll();

		TestDao result = injector.get(TestDao.class, "testDaoAnother");

		assertThat(result, sameInstance((TestDao) testDaoAnother));
	}

	@Test
	public void set_상위클래스_필드() {
		injector.set(testParentDao);

		assertThat(testService.getTestParentDao(), sameInstance(testParentDao));
	}

	@Test
	public void get_상위클래스_필드() {
		injectAll();

		TestParentDao result = injector.get(TestParentDao.class);

		assertThat(result, sameInstance(testParentDao));
	}
}